@@ -1,4 +1,4 @@
-#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -20,8 +20,8 @@
 """
 
 version_string = """\
-Nuitka V0.6.17.7
-Copyright (C) 2021 Kay Hayen."""
+Nuitka V1.1.6
+Copyright (C) 2022 Kay Hayen."""
 
 
 def getNuitkaVersion():
@@ -32,6 +32,32 @@ def getNuitkaVersion():
     return version_string.split()[1][1:]
 
 
+def getNuitkaVersionTuple():
+    """Return Nuitka version as a string.
+
+    This can also not be used for precise comparisons, last one might contain "rc"
+    """
+
+    version = getNuitkaVersion()
+
+    if "rc" in version:
+        rc_number = int(version[version.find("rc") + 2 :] or "0")
+        version = version[: version.find("rc")]
+
+        is_final = False
+    else:
+        rc_number = 0
+        is_final = True
+
+    result = version.split(".")
+    if len(result) == 2:
+        result.append("0")
+
+    result = [int(digit) for digit in result]
+    result.extend((is_final, rc_number))
+    return tuple(result)
+
+
 def getNuitkaVersionYear():
     """The year of Nuitka copyright for use in generations."""
 
